from enum import Enum
import chainer
from chainer.optimizers import MomentumSGD, Adam, RMSprop
from chainer.training import extensions

class OptimizerType(Enum):
	SGD = 1
	RMSPROP = 2
	ADAM = 3
	Default = SGD

	@classmethod
	def as_choices(cls):
		return {e.name.lower(): e for e in cls}

	@classmethod
	def get(cls, key):
		key = key.lower()
		choices = cls.as_choices()
		return choices.get(key) if key in choices else cls.Default

def optimizer(opt_type_name, model, lr=1e-2, decay=9e-1, *args, **kw):
	opt_type = OptimizerType.get(opt_type_name)
	opt_cls = {
		OptimizerType.SGD: MomentumSGD,
		OptimizerType.RMSPROP: RMSprop,
		OptimizerType.ADAM: Adam,
	}.get(opt_type)

	opt_args = dict(alpha=lr) if opt_type == OptimizerType.ADAM else dict(lr=lr)
	kw.update(opt_args)
	opt = opt_cls(*args, **kw)
	opt.setup(model)
	if decay:
		opt.add_hook(chainer.optimizer.WeightDecay(decay))

	return opt

def lr_shift(opt, init, rate, target):
	attr = "alpha" if isinstance(opt, Adam) else "lr"

	return extensions.ExponentialShift(
		optimizer=opt, attr=attr,
		init=init, rate=rate, target=target)

def no_grad(arr):
	return arr.data if isinstance(arr, chainer.Variable) else arr




def all_children(model):
	"""
		Iterate over all links in a (nested) chain.
	"""

	for child in model.children():
		grand_children = list(child.children())
		if grand_children:
			for grand_child in all_children(child):
				yield grand_child
		else:
			yield child



#!/usr/bin/env python
if __name__ != '__main__': raise Exception("Do not import me!")

"""
Script for plotting loss and accuracy from chainer logs

usage:
	> python -m chainer_addon.utils.plot <logfile> [-k <key1> <key2> ... <keyn> ]
"""

import sys, argparse

parser = argparse.ArgumentParser(description = "Plot histograms of all parameters in given data")
parser.add_argument("data", nargs="+", type=str, help="data to visualize")
parser.add_argument("--pattern", type=str, help="pattern to sort out parameter names")
parser.add_argument("--n_bins", type=int, default=10, help="number of bins for the histograms")

from os.path import join, basename
from scipy.optimize import curve_fit
import simplejson as json, numpy as np, re
import matplotlib.pyplot as plt

def func(x, a, b, c):
	return a * np.exp(-b * x) + c

def non_zero(idxs, values):
	return list(zip(*[(idx, val) for idx, val in zip(idxs, values) if val != 0]))

def main(args):
	models = list(map(np.load, args.data))
	pattern_regex = None
	if args.pattern:
		pattern_regex = re.compile(args.pattern)

	# gather matching keys
	keys = []
	for key in models[0].keys():
		if pattern_regex is not None and not re.findall(pattern_regex, key): continue
		keys.append(key)

	if not keys:
		print("No parameters were selected!", file=sys.stderr)
		exit()

	keys.sort()

	cols = int(np.ceil(np.sqrt(len(keys))))
	rows = int(np.ceil(len(keys) / cols))
	fig = plt.figure()

	print("Plotting \n\t{}".format(", ".join(keys)))
	for i, key in enumerate(keys, 1):
		ax = fig.add_subplot(rows, cols, i)
		ax.set_title(key)
		ax.grid()

		for name, model in zip(args.data, models):
			param = model[key]
			histo, X = np.histogram(param, density=True, bins=args.n_bins)
			x = [(x2 + x1) / 2 for x1, x2 in zip(X[:-1], X[1:])]
			ax.plot(x, histo, label=name)
		ax.legend()

	plt.show()

main(parser.parse_args())





import numpy as np
import chainer.links as L
import chainer.functions as F

from chainer.serializers import npz
from chainer.links.model.vision.resnet import prepare

from chainer_addons.models import PretrainedModelMixin

from collections import OrderedDict

class ResnetLayers(PretrainedModelMixin, L.ResNet50Layers):

	class meta:
		classifier_layers = ["fc6"]
		conv_map_layer = "res5"
		feature_layer = "pool5"
		feature_size = 2048
		n_conv_maps = 2048
		input_size = 448
		mean = np.array([103.063,  115.903,  123.152], dtype=np.float32).reshape(3,1,1)

	@staticmethod
	def prepare(*args, **kw):
		return prepare(*args, **kw)

	@staticmethod
	def pooling(x):
		n, channel, rows, cols = x.data.shape
		h = F.average_pooling_2d(x, (rows, cols), stride=1)
		h = F.reshape(h, (n, channel))
		return h

	def __init__(self, pretrained_model, n_classes):
		if pretrained_model == "auto":
			super(ResnetLayers, self).__init__(pretrained_model=pretrained_model)
		else:
			super(ResnetLayers, self).__init__(pretrained_model=None)

		with self.init_scope():
			delattr(self, "fc6")
			self.fc6 = L.Linear(2048, n_classes)

		if pretrained_model not in [None, "auto"]:
			npz.load_npz(pretrained_model, self)

	@property
	def functions(self):
		links = [
				('conv1', [self.conv1, self.bn1, F.relu]),
				('pool1', [F.MaxPooling2D(ksize=3, stride=2)]),
				('res2', [self.res2]),
				('res3', [self.res3]),
				('res4', [self.res4]),
				('res5', [self.res5]),
				('pool5', [ResnetLayers.pooling])]
		if hasattr(self, "fc6"):
			links +=[
				('fc6', [self.fc6]),
				('prob', [F.softmax]),
			]

		return OrderedDict(links)

